import { MathUtil } from "./math.js";
import { MenuSelection } from "./menu_selection.js";
import { Point } from "./point.js";
import { Rect } from "./rect.js";
export var Preview;
(function (Preview) {
    const MANIPULATION_SCALE = 2.2;
    const MINIMUM_SIZE = 20;
    const ONE_THIRD = 1.0 / 3.0;
    const TWO_THIRD = 2.0 / 3.0;
    function oppose(loc) {
        return (loc + 4) % 8; // Magic!
    }
    Preview.oppose = oppose;
    class LayoutPreview {
        Sys;
        size;
        base;
        preview = null;
        selection;
        windowRect;
        win;
        visible;
        ui;
        resizeCorner;
        trackingOrigin;
        constructor(Sys, size, windowRect, win) {
            this.Sys = Sys;
            this.size = size;
            this.ui = Sys.newClutterActor();
            this.ui.set_background_color(Sys.newClutterColor({
                red: 80,
                green: 158,
                blue: 255,
                alpha: 125
            }));
            this.resizeCorner = null;
            this.selection = MenuSelection.None(null);
            this.windowRect = windowRect;
            this.win = win;
            this.ui.hide();
            this.visible = false;
        }
        selectOuter(loc, splitMode) {
            const size = this.size;
            switch (splitMode) {
                case 0 /* SplitMode.FOUR */:
                    switch (loc) {
                        case 0 /* Anchor.LEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: 0.5, y: 1 }, size),
                            };
                        case 1 /* Anchor.TOPLEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scaleConstant(0.5, size),
                            };
                        case 2 /* Anchor.TOP */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: 1, y: 0.5 }, size),
                            };
                        case 3 /* Anchor.TOPRIGHT */:
                            return {
                                pos: Point.scale({ x: 0.5, y: 0 }, size),
                                size: Point.scaleConstant(0.5, size),
                            };
                        case 4 /* Anchor.RIGHT */:
                            return {
                                pos: Point.scale({ x: 0.5, y: 0 }, size),
                                size: Point.scale({ x: 0.5, y: 1 }, size),
                            };
                        case 5 /* Anchor.BOTTOMRIGHT */:
                            return {
                                pos: Point.scaleConstant(0.5, size),
                                size: Point.scaleConstant(0.5, size),
                            };
                        case 6 /* Anchor.BOTTOM */:
                            return {
                                pos: Point.scale({ x: 0, y: 0.5 }, size),
                                size: Point.scale({ x: 1, y: 0.5 }, size),
                            };
                        case 7 /* Anchor.BOTTOMLEFT */:
                            return {
                                pos: Point.scale({ x: 0, y: 0.5 }, size),
                                size: Point.scaleConstant(0.5, size),
                            };
                    }
                case 1 /* SplitMode.SIX */:
                    switch (loc) {
                        case 0 /* Anchor.LEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: ONE_THIRD, y: 1 }, size),
                            };
                        case 1 /* Anchor.TOPLEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                            };
                        case 2 /* Anchor.TOP */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                            };
                        case 3 /* Anchor.TOPRIGHT */:
                            return {
                                pos: Point.scale({ x: TWO_THIRD, y: 0 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                            };
                        case 4 /* Anchor.RIGHT */:
                            return {
                                pos: Point.scale({ x: TWO_THIRD, y: 0 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 1 }, size),
                            };
                        case 5 /* Anchor.BOTTOMRIGHT */:
                            return {
                                pos: Point.scale({ x: TWO_THIRD, y: 0.5 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                            };
                        case 6 /* Anchor.BOTTOM */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                            };
                        case 7 /* Anchor.BOTTOMLEFT */:
                            return {
                                pos: Point.scale({ x: 0, y: 0.5 }, size),
                                size: Point.scale({ x: ONE_THIRD, y: 0.5 }, size),
                            };
                    }
            }
            return null;
        }
        selectInner(sel, splitMode) {
            switch (splitMode) {
                case 0 /* SplitMode.FOUR */:
                    switch (sel) {
                        case 2 /* Anchor.TOP */:
                            return {
                                pos: Point.ZERO,
                                size: this.size,
                            };
                        case 6 /* Anchor.BOTTOM */:
                            return null;
                    }
                    break;
                case 1 /* SplitMode.SIX */:
                    switch (sel) {
                        case 1 /* Anchor.TOPLEFT */:
                            return {
                                pos: Point.ZERO,
                                size: Point.scale({ x: TWO_THIRD, y: 1 }, this.size),
                            };
                        case 3 /* Anchor.TOPRIGHT */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0 }, this.size),
                                size: Point.scale({ x: TWO_THIRD, y: 1 }, this.size),
                            };
                        case 6 /* Anchor.BOTTOM */:
                            return {
                                pos: Point.scale({ x: ONE_THIRD, y: 0 }, this.size),
                                size: Point.scale({ x: ONE_THIRD, y: 1 }, this.size),
                            };
                    }
                    break;
            }
            return null;
        }
        trackMouse(prevMode, origin) {
            if (prevMode == 0 /* MouseMode.MENU */) {
                switch (this.selection.ring) {
                    case 1 /* Ring.INNER */:
                        return false;
                    case 2 /* Ring.OUTER */:
                        this.resizeCorner = oppose(this.selection.index);
                        this.trackingOrigin = origin;
                        break;
                    case 0 /* Ring.NONE */:
                        this.resetTracking(origin);
                        break;
                }
            }
            else {
                this.resetTracking(origin);
            }
            this.base = this.preview; // capture whatever we have as a base
            this.updateUi();
            return true;
        }
        resetTracking(origin) {
            if (this.preview == null) {
                this.preview = this.windowRect;
            }
            this.base = this.preview;
            this.resizeCorner = Rect.closestCorner(this.preview, origin);
            this.trackingOrigin = origin;
            this.updateUi();
        }
        onMouseMove(mode, point) {
            switch (mode) {
                case 1 /* MouseMode.RESIZE */:
                    if (this.resizeCorner === null) {
                        return;
                    }
                    var diff = Point.subtract(point, this.trackingOrigin);
                    this.preview = LayoutPreview.applyResize(this.resizeCorner, diff, this.base, this.size);
                    // p('move diff ' + JSON.stringify(diff)
                    // 	+ ' (from origin ' + JSON.stringify(this.trackingOrigin) + ')'
                    // 	+ ' turned base ' + JSON.stringify(this.base)
                    // 	+ ' into rect ' + JSON.stringify(this.preview)
                    // );
                    break;
                case 2 /* MouseMode.MOVE */:
                    var diff = Point.subtract(point, this.trackingOrigin);
                    this.preview = LayoutPreview.applyMove(diff, this.base, this.size);
                    // p('move diff ' + JSON.stringify(diff)
                    // 	+ ' (from origin ' + JSON.stringify(this.trackingOrigin) + ')'
                    // 	+ ' turned base ' + JSON.stringify(this.base)
                    // 	+ ' into rect ' + JSON.stringify(this.preview)
                    // );
                    break;
                default:
                    return;
            }
            this.updateUi();
        }
        applyManipulator(fn) {
            if (this.preview === null)
                return;
            this.preview = fn(this.preview, this.size);
            this.updateUi();
        }
        static applyMove(diff, base, bounds) {
            const ret = Rect.copy(base);
            const scaled = Point.scaleConstant(MANIPULATION_SCALE, diff);
            ret.pos.x = Math.max(0, Math.min(ret.pos.x + scaled.x, bounds.x - ret.size.x));
            ret.pos.y = Math.max(0, Math.min(ret.pos.y + scaled.y, bounds.y - ret.size.y));
            return ret;
        }
        static applyResize(location, diff, base, bounds) {
            const ret = Rect.copy(base);
            const scaled = Point.scaleConstant(MANIPULATION_SCALE, diff);
            function moveNear(axis) {
                // minimum diff is enough to bring this edge to 0 (i.e. invert the current pos)
                // maximum diff is enough to bring this edge to ther other side of this rect, minus MINIMUM_SIZE
                // p('moveNear['+axis+']');
                const diff = MathUtil.between(-ret.pos[axis], scaled[axis], ret.size[axis] - MINIMUM_SIZE);
                ret.pos[axis] += diff;
                ret.size[axis] -= diff;
            }
            function moveFar(axis) {
                // minimum diff is enough to bring this edge to the other side of this rect, plus MINIMUM_SIZE
                // maximum diff is enough to bring this edge to the right bounds
                // p('moveFar['+axis+']');
                const diff = MathUtil.between(MINIMUM_SIZE - ret.size[axis], scaled[axis], bounds[axis] - ret.pos[axis] - ret.size[axis]);
                ret.size[axis] += diff;
            }
            switch (location) {
                case 0 /* Anchor.LEFT */:
                    moveNear("x" /* Axis.x */);
                    break;
                case 1 /* Anchor.TOPLEFT */:
                    moveNear("x" /* Axis.x */);
                    moveNear("y" /* Axis.y */);
                    break;
                case 2 /* Anchor.TOP */:
                    moveNear("y" /* Axis.y */);
                    break;
                case 3 /* Anchor.TOPRIGHT */:
                    moveNear("y" /* Axis.y */);
                    moveFar("x" /* Axis.x */);
                    break;
                case 4 /* Anchor.RIGHT */:
                    moveFar("x" /* Axis.x */);
                    break;
                case 5 /* Anchor.BOTTOMRIGHT */:
                    moveFar("x" /* Axis.x */);
                    moveFar("y" /* Axis.y */);
                    break;
                case 6 /* Anchor.BOTTOM */:
                    moveFar("y" /* Axis.y */);
                    break;
                case 7 /* Anchor.BOTTOMLEFT */:
                    moveNear("x" /* Axis.x */);
                    moveFar("y" /* Axis.y */);
                    break;
                default:
                    throw new Error("unknown location: " + location);
            }
            return ret;
        }
        updateSelection(sel) {
            this.selection = sel;
            switch (sel.ring) {
                case 2 /* Ring.OUTER */:
                    this.base = this.selectOuter(sel.index, sel.splitMode);
                    break;
                case 1 /* Ring.INNER */:
                    this.base = this.selectInner(sel.index, sel.splitMode);
                    break;
                case 0 /* Ring.NONE */:
                default:
                    this.base = null;
                    break;
            }
            this.resetPreview();
        }
        setVisible(visible) {
            // Whenever the preview is visible, the window is hidden
            if (this.visible == visible) {
                return;
            }
            // p('setting visible to: ' + visible)
            if (this.win !== null) {
                this.Sys.setWindowHidden(this.win, visible);
            }
            if (visible) {
                this.ui.show();
            }
            else {
                this.ui.hide();
            }
            this.visible = visible;
        }
        updateUi() {
            if (this.preview == null) {
                this.setVisible(false);
            }
            else {
                this.ui.set_size(this.preview.size.x, this.preview.size.y);
                this.ui.set_position(this.preview.pos.x, this.preview.pos.y);
                this.setVisible(true);
            }
        }
        resetPreview() {
            if (this.base == null) {
                this.preview = null;
            }
            else {
                this.preview = Rect.copy(this.base);
            }
            this.updateUi();
        }
        getRect() { return this.preview; }
    }
    Preview.LayoutPreview = LayoutPreview;
})(Preview || (Preview = {}));
